/** \file
 * 
 * Sep 2, 2013
 *
 * Copyright Ian Kaplan 2013
 *
 * @author Ian Kaplan, www.bearcave.com, iank@bearcave.com
 */
package t9;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.net.URL;
import java.util.LinkedHashSet;

import t9.trie.Pair;
import t9.trie.TrieBuilder;
import t9.trie.TrieNode;

/**
 * Dictionary
 * Sep 4, 2013
 * 
 * @author Ian Kaplan, iank@bearcave.com
 */
public class Dictionary {
    private final static String wordsFile = "5000_common_words.csv";
    
    private static Dictionary instance = null;
    private static TrieNode[] mDictionary;
    
    private String getPathForResource( String filename )
    {
        ClassLoader loader = this.getClass().getClassLoader();
        URL url = loader.getResource( filename );
        String path = null;
        if (url != null) {
            path = url.getFile();
        }
        return path;
    } // getPathForResource
    
    /**
     * Initialize the Dictionary by reading a file that contains comma separated
     * words and relative frequency. This code depends on the entries in the file
     * being ordered since a set of prefix Tries will be created where each prefix
     * Trie starts with the same character.
     * 
     * @throws FileNotFoundException
     */
    protected Dictionary() throws FileNotFoundException, IOException {
        int numTries = ((int)'z' - (int)'a') + 1;
        mDictionary = new TrieNode[ numTries ];
        String wordPath = getPathForResource( wordsFile );
        File wordFile = new File( wordPath );
        if (wordFile.exists() && wordFile.canRead()) {
            FileReader reader = new FileReader( wordFile );
            BufferedReader bufReader = null;
            try {
                bufReader = new BufferedReader( reader );
                String line = null;
                char startChar = 'a';
                LinkedHashSet<Pair> set = new LinkedHashSet<Pair>();
                TrieBuilder builder = new TrieBuilder();
                int lineCnt = 0;
                while ((line = bufReader.readLine()) != null) {
                    lineCnt++;
                    String[] parts = line.split(",");
                    if (parts.length == 2) {
                        String word = parts[0].toLowerCase();
                        int freq = Integer.parseInt( parts[1] );
                        Pair pair = new Pair(word, freq );
                        if (word.charAt(0) == startChar) {
                            set.add(pair);
                        }
                        else {
                            TrieNode root = builder.buildTrie( set );
                            int ix = (int)startChar - (int)'a';
                            mDictionary[ix] = root;
                            set.clear();
                        }
                    }
                    else {
                        throw new IOException("Bad format, in file " + wordsFile + " line " + lineCnt );
                    }
                } // while
            }
            finally {
                if (bufReader != null) { bufReader.close(); }
            }
        }
        else {
            final String msg = "Could not read " + wordsFile; 
            throw new FileNotFoundException( msg );
        }
    } // Dictionary constructor

    /**
     * Get an instance of the Dictionary singleton.
     * 
     * @return an instance of the Dictionary singleton
     * @throws FileNotFoundException 
     * @throws IOException 
     */
    public static Dictionary getDictionary() throws FileNotFoundException, IOException
    {
        if (instance == null) { 
            instance = new Dictionary();
        }
        return instance;
    }
        
    /**
     * Get the prefix tree associated with a starting letter.
     * 
     * @param letter the starting letter for the prefix Trie
     * @return the root of the prefix tree or null if for some reason the Trie doesn't exist.
     */
    public TrieNode getPrefixTrie(final char letter) {
        char l = Character.toLowerCase(letter);
        TrieNode root = null;
        if (Character.isLetter(l)) {
            int ix = (int)l - (int)'z';
            root = mDictionary[ ix ];
        }
        return root;
    }
}
