/** \file
 * 
 * <p>
 * A string prefix "Trie", which also contains the relative frequency for each word in the Trie.
 * </p>
 * <p>
 * According to "Algorithms in Java", Third Edition by Robert Sedgewick the Trie was first
 * "discovered" by de la Briandais in 1959 and given the name Trie by Fredkin in 1960.
 * </p>
 * <p>
 * The Trie algorithm for this class is of my own design.  I looked for other algorithms but didn't
 * find an algorithm that was memory efficient (several algorithms I found replaced the sibling list
 * with a vector with an element for each letter, instead of the sibling prefix list I use).
 * </p>
 * <p>
 * It is important to note that the
 * Trie is ordered on the basis of the letters that make up the words that are stored in the
 * Trie. Although the sibling lists are ordered, the is no other lexical ordering in the tree.
 * </p>
 * <p>
 * The prefix Trie is designed for words that share the same first character and the root of the
 * Trie is this character.
 * </p>
 * 
 * Sep 2, 2013
 *
 * Copyright Ian Kaplan 2013
 *
 * @author Ian Kaplan, www.bearcave.com, iank@bearcave.com
 */
package t9.trie;

import java.util.Collection;
import java.util.TreeSet;

/**
 * TrieBuilder
 * Sep 3, 2013
 * 
 * <p>
 * Support for the construction of a String prefix "trie". In this case the prefix tree has a single
 * root, which will be the letter that starts all of the words in the tree. The tree is a directed
 * graph, but it is not ordered in the way a binary tree is. Instead the tree is ordered by the
 * structure of the words.
 * </p>
 * <p>
 * A tree containing the words:
 * </p>
 * <pre>
 * a
 * abandon
 * able
 * abuse
 * academic
 * accept
 * acceptable
 * </pre>
 * <pre>
 *    child ===>
 *    sibling
 *      |
 *      |
 *      V
 *      
 *    a - b - a - n - d - o - n
 *    *   |   |               *
 *        |   l - e
 *        |   |   *
 *        |   |
 *        |   u - s - e
 *        |           *
 *        |
 *        c - a - d - e - m - i - c
 *            |                   *
 *            c - e - p - t
 *                        |*
 *                        a - b - l - e
 *            
 * </pre>
 * <p>
 * The words are ordered in the child direction. The siblings are the 
 * words that have the same prefix. The star indicates an end-of-word.
 * </p>
 * <p>
 * Note that the sibling words are lexically ordered.
 * </p>
 * <p>
 * Each word has an associated relative frequency that annotates the last character
 * in the word. Relative frequency is the frequency ordering within a corpus. For example,
 * the word "a" is the 5th most frequent word out of a set of 5000 words.
 * </p>
 * @author Ian Kaplan, iank@bearcave.com
 */
public class TrieBuilder {

    /**
     * <p>
     * Insert a new "sibling" character into the ordered sibling list.
     * </p>
     * <p>
     * Words are read down the child links. The sibling links connect to 
     * other words with the same prefix (not that the prefix also runs down the
     * child links).
     * </p>
     * <p>
     * For example:
     * </p>
     * <pre>
     * words: 
     *    aardvark
     *    able
     *    action
     *    
     *    child ===>
     *    sibling
     *      |
     *      |
     *      V
     *    
     * a - a - r - d - v - a - r - k
     *     |
     *     b - l - e
     *     |
     *     c - t - i - o - n
     * </pre>
     * @param sibling the current head of a list of sibling words that share the same prefix
     * @param letter the letter to be found or inserted into the sibling list.
     * @return the part of the tree that contains the character which was either found or 
     *         inserted.
     */
    private TrieNode listInsert(TrieNode sibling, char letter) {
        TrieNode head = sibling;
        TrieNode trail = head;
        while (head != null) {
            if (letter == head.getLetter()) {
                break;
            }
            else if (letter < head.getLetter()) {
                TrieNode newSib = new TrieNode(letter);
                trail.setSib(newSib);
                newSib.setSib(head);
                head = newSib;
                break;
            }
            trail = head;
            head = head.getSib();
            if (head == null) {
                TrieNode newSib = new TrieNode(letter );
                trail.setSib(newSib);
                head = newSib;
            }
        }
        return head;
    }
 

    /**
     * Insert a character into the sibling list of words that have the same prefix
     * @param head a reference to the TrieNode which has or will have the sibling list
     * @param letter the letter to be found or inserted.
     * @return the TrieNode for the character that was either found or inserted.
     */
    private TrieNode insertSibling(TrieNode head, char letter) {
        TrieNode sibling = head.getSib();
        if (sibling == null) {
            TrieNode newSib = new TrieNode(letter);
            head.setSib(newSib);
            head = newSib;
        }
        else {
            head = listInsert(sibling, letter );
        }
        return head;
    }



    /**
     * Iterate over the characters in a word and add them to the Trie if they are not
     * already there.
     * 
     * @param root the root of the Trie
     * @param a Pair container that contains the String to be inserted and the relative frequency value.
     * @return the root of the prefix trie
     */
    public TrieNode addWord(TrieNode root, Pair word) {
        if (word != null) {
            TrieNode head = root;
            TrieNode trail = head;
            char letters[] = word.word.toCharArray();
            for (char letter : letters) {
                if (root == null) {
                    root = new TrieNode( letter );
                    trail = root;
                }
                else if (head == null) {
                    TrieNode newChild = new TrieNode(letter);
                    trail.setChild(newChild);
                    trail = newChild;
                }
                else if (head.getLetter() != letter){
                    head = insertSibling( head, letter );
                    trail = head;
                    head = head.getChild();
                }
                else {
                    trail = head;
                    head = head.getChild();
                }
            }
            if (head == null) {
                head = trail;
            }
            head.setEOW();
            head.setFrequency(word.frequency);
        }

        return root;
    }  

    /**
     * Add a collecton of Pair objects to the prefix Trie.
     * 
     * @param words a list of sorted words that all start with the same letter
     * @return the root of the trie constructed for the words
     */
    public TrieNode buildTrie( Collection<Pair> words ) {
        TrieNode root = null;
        for (Pair word : words) {
            root = addWord(root, word);
        }
        return root;
    }

    /**
     * Return a Collection of Pair objects that consist of the words in the tree and the
     * relative frequency for each word.
     * 
     * @param root
     * @param prefix
     * @param words
     */
    protected void listTrie( TrieNode root, String prefix, Collection<Pair> words ) {
        String newPrefix = prefix;
        for (TrieNode curRoot = root; curRoot != null; curRoot = curRoot.getChild()) {
            // traverse down the sibling list
            TrieNode sib = curRoot.getSib();
            listTrie(sib, newPrefix, words );
            newPrefix = newPrefix + curRoot.getLetter();
            if (curRoot.isEOW()) {
                Pair pair = new Pair(newPrefix, curRoot.getFrequency());
                words.add(pair);
            }
        }
    }
    
    /**
     * Return the contents of the Trie as a list of words and frequency values.
     * 
     * @param root the root of the prefix Trie
     * @return a sorted TreeSet containing the Pair objected lexically sorted by the String
     *         values.
     */
    public TreeSet<Pair> listTrie(TrieNode root)
    {
        TreeSet<Pair> words = new TreeSet<Pair>();
        listTrie(root, "", words);
        return words;
    }

}
